from functools import wraps, partial
from typing import Callable
from flask import request
from src.errors import RequestValidationError
from marshmallow import ValidationError

    
def request_valiator(f: Callable, body_schema_class=None, query_schema_class=None, param_schema_class=None):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        try:
            if body_schema_class is not None:
                body_data = request.get_json()
                body_schema_class().load(data = body_data)
            if query_schema_class is not None:
                query_data = request.args.to_dict()
                print(query_data)
                query_schema_class().load(data = query_data)
            if param_schema_class is not None:
                param_data = request.view_args
                param_schema_class().load(data = param_data)
        except ValidationError as e:
            errors = [{ "param": value, "msg": e.messages[value][0]} for key, value in enumerate(e.messages) ]
            raise RequestValidationError(errors=errors)
        return f(*args, **kwargs)
    return decorated_function

def validate_request (body_schema_class=None, query_schema_class=None, param_schema_class=None):
    return partial(request_valiator, body_schema_class=body_schema_class, query_schema_class=query_schema_class, param_schema_class=param_schema_class)


from .custom_error import CustomError, ExceptionMessage
from typing import List, TypedDict

class ValidationDict(TypedDict):
    msg: str
    param: str
class RequestValidationError(CustomError):
    status_code: int = 400
    errors: List[ValidationDict]

    def __init__(self, errors: List[ValidationDict]) -> None:
        self.errors = errors
        super().__init__('Invalid Request Parameters')
    
    def serialize_errors(self) -> List[ExceptionMessage]:
        return list(map(lambda error: {"message": error["msg"] , "field": error["param"] }, self.errors))
        


from .custom_error import CustomError, ExceptionMessage
from typing import List

class BadRequestError(CustomError):
    status_code: int = 400

    def __init__(self, message: str) -> None:
        super().__init__(message)
    
    def serialize_errors(self) -> List[ExceptionMessage]:
        return [{"message": self.message}]        


from abc import ABC, abstractproperty
from typing import List, TypedDict, Optional

class ExceptionMessage(TypedDict):
    message: str
    field: Optional[str]

class CustomError(ABC, Exception):
    def status_code() -> int: pass
    def message()-> str: pass 

    def __init__(self, message:str) -> None:
        self.message = message
        super().__init__(message)
    
    @abstractproperty
    def serialize_errors(self) -> List[ExceptionMessage]:
        pass


import enum
from typing import Literal, get_args
import typing

class Subjects(enum.Enum):
    UserCreated = 'user:created'

SubjectsValues = Literal['user:created']
assert set(typing.get_args(SubjectsValues)) == {member.value for member in Subjects}


from .custom_error import CustomError, ExceptionMessage
from typing import List

class NotFoundError(CustomError):
    status_code: int = 404

    def __init__(self) -> None:
        super().__init__('Not Found')
    
    def serialize_errors(self) -> List[ExceptionMessage]:
        return [{"message": self.message}] 


from flask import jsonify
from src.errors import CustomError

def exception_handler(e: Exception):
    if isinstance(e, CustomError):
        return jsonify({ "errors": e.serialize_errors()}), e.status_code
    print(e)
    return jsonify({"errors": [{"message": "Something went wrong"}]}), 400


from typing import Any
from flask import Flask
from werkzeug.wrappers import Request
import base64, json

class NodeCookieSessionDecode:
    def __init__(self, app:Flask):
        self.app = app
    
    def __call__(self, environ, start_response, *args: Any, **kwargs: Any) -> Any:
        request = Request(environ)
        try:
            session_cookie = request.cookies.get('session')
            session = json.loads(base64.b64decode(session_cookie)) or {}
            environ['session'] = session
        except:
            pass
        return self.app(environ, start_response)



from functools import wraps
from typing import Callable
from flask import request
from src.errors import NotAuthorizedError

def require_auth(f: Callable):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        if not hasattr(request, 'current_user') or request.current_user is None:
            raise NotAuthorizedError()
        return f(*args, **kwargs)
    return decorated_function


from .cookie_session import NodeCookieSessionDecode
from .current_user import current_user
from .exception_handler import exception_handler
from .require_auth import require_auth
from .validate_request import validate_request



import os
from typing import Any, List, Optional, TypedDict, Callable
import jwt
from src.types import UserRoleValues, SubscriptionTypeValues
# from functools import wraps
from flask import request

class Subscription(TypedDict):
    id: str
    subscriptionType: SubscriptionTypeValues
    expiryDate: str

class UserPayload(TypedDict):
    id: str
    email: str
    roles: List[UserRoleValues]
    firstName: Optional[str]
    lastName: Optional[str]
    active: bool
    banned: bool
    subscription: Optional[Subscription]
    billingId: Optional[str]
    mobileNumber: Optional[str]
    loginCount: int
    lastLogin: str
    profileImageUrl: Optional[str]

def current_user():    
    try:
        userJWT = request.environ['session']['jwt']
        current_user = jwt.decode(userJWT, os.environ.get('JWT_KEY'), ['HS256']) 
    except:
        current_user = None
    setattr(request, 'current_user', current_user)
        


from .bad_request_error import BadRequestError
from .custom_error import CustomError
from .not_authorized_error import NotAuthorizedError
from .not_found_error import NotFoundError
from .request_validation_error import RequestValidationError


import enum
from typing import Literal, get_args

class SubscriptionType(enum.Enum):
    BASIC = 'BASIC'
    PRO = 'PRO'
    PREMIUM = 'PREMIUM'

SubscriptionTypeValues = Literal['BASIC', 'PRO', 'PREMIUM']
assert set(get_args(SubscriptionTypeValues)) == {member.value for member in SubscriptionType}


import enum
from typing import Literal, get_args

class UserRoles(enum.Enum):
    Normal = 'normal'
    Moderator = 'moderator'
    Developer = 'developer'
    Admin = 'admin'

UserRoleValues = Literal['normal', 'moderator', 'developer', 'admin']
assert set(get_args(UserRoleValues)) == {member.value for member in UserRoles}
